package edu.berkeley.qfat;
import java.awt.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.media.opengl.*;
import javax.media.opengl.glu.*;
import edu.berkeley.qfat.geom.*;
import edu.wlu.cs.levy.CG.KDTree;
import edu.berkeley.qfat.geom.Point;

public class Mesh implements Iterable<Mesh.T> {

    public static final float EPSILON = (float)0.0001;
    public static final Random random = new Random();

    private RTree<T>         triangles = new RTree<T>();
    private PointSet<Vertex> vertices  = new PointSet<Vertex>();

    public boolean immutableVertices;
    public boolean ignorecollision = false;
    public Mesh    score_against = null;
    public double  score = 0;

    public Mesh(boolean immutableVertices) { this.immutableVertices = immutableVertices; }

    public void makeVerticesImmutable() { this.immutableVertices = true; }
    public float score() { return (float)score; }

    public int size() { return vertices.size(); }
    public Iterable<Vertex> vertices() { return vertices; }
    public Iterator<T> iterator() { return triangles.iterator(); }

    public void rebindPoints() {
        // unbind all points
        for(Mesh.T t : this) {
            t.v1().unbind();
            t.v2().unbind();
            t.v3().unbind();
        }
        // ask edges to re-implement their bindings
        for(Mesh.T t : this) {
            t.e1().dobind();
            t.e2().dobind();
            t.e3().dobind();
        }
    }

    public void unApplyQuadricToNeighborAll() {
        HashSet<Vertex> done = new HashSet<Vertex>();
        for(T t : this)
            for(Vertex p : new Vertex[] { t.v1(), t.v2(), t.v3() }) {
                if (done.contains(p)) continue;
                done.add(p);
                p.unApplyQuadricToNeighbor();
            }
    }
    public void recomputeAllFundamentalQuadrics() {
        HashSet<Vertex> done = new HashSet<Vertex>();
        for(T t : this)
            for(Vertex p : new Vertex[] { t.v1(), t.v2(), t.v3() }) {
                if (done.contains(p)) continue;
                done.add(p);
                p.recomputeFundamentalQuadric();
            }
    }
    public float applyQuadricToNeighborAll() {
        int num = 0;
        double dist = 0;
        HashSet<Vertex> done = new HashSet<Vertex>();
        for(T t : this)
            for(Vertex p : new Vertex[] { t.v1(), t.v2(), t.v3() }) {
                if (done.contains(p)) continue;
                done.add(p);
                p.applyQuadricToNeighbor();
                
            }
        return (float)(dist/num);
    }

    public void transform(Matrix m) {
        ArrayList<Vertex> set = new ArrayList<Vertex>();
        for(Vertex v : vertices) set.add(v);
        for(Vertex v : set) v.transform(m);
    }

    public void rebuild() { /*vertices.rebuild();*/ }
    public Vec diagonal() { return vertices.diagonal(); }
    public Point centroid() { return vertices.centroid(); }
    public Vertex nearest(Point p) { return vertices.nearest(p); }

    /** compute the volume of the mesh */
    public float volume() {
        double total = 0;
        for(T t : this) {
            double area = t.area();
            Vec origin_to_centroid = new Vec(new Point(0, 0, 0), t.centroid());
            boolean facingAway = t.norm().dot(origin_to_centroid) > 0;
            double height = Math.abs(t.norm().dot(origin_to_centroid));
            total += ((facingAway ? 1 : -1) * area * height) / 3.0;
        }
        return (float)total;
    }


    // Vertexices //////////////////////////////////////////////////////////////////////////////

    /** a vertex in the mesh */
    public final class Vertex extends HasPoint implements Visitor<T> {
        public String toString() { return p.toString(); }
        public Point p;
        E e;                // some edge *leaving* this point

        /** the nearest vertex in the "score_against" mesh */
        Vertex   nearest_in_other_mesh;
        /** the number of vertices in the other mesh for which this is the nearest_in_other_mesh */
        int    quadric_count;
        /** the total error quadric (contributions from all vertices in other mesh for which this is nearest) */
        Matrix quadric = Matrix.ZERO;

        Vertex bound_to = this;
        Matrix binding = Matrix.ONE;
        float oldscore = 0;
        boolean quadricStale = false;

        public Matrix errorQuadric() { return quadric; }
        public Point getPoint() { return p; }
        public float score() { return oldscore; }

        private Matrix fundamentalQuadric = null;
        public Matrix fundamentalQuadric() {
            if (fundamentalQuadric == null) recomputeFundamentalQuadric();
            return fundamentalQuadric;
        }

        private Vertex(Point p) {
            this.p = p;
            if (vertices.get(p) != null) throw new Error();
            vertices.add(this);
        }

        private void glNormal(GL gl) {
            Vec norm = norm();
            gl.glNormal3f(norm.x, norm.y, norm.z);
        }

        public void recomputeFundamentalQuadric() {
            if (!quadricStale && fundamentalQuadric != null) return;
            quadricStale = false;
            unApplyQuadricToNeighbor();
            Matrix m = Matrix.ZERO;
            int count = 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                T t = e.t;
                m = m.plus(t.norm().fundamentalQuadric(t.centroid()));
                count++;
            }
            fundamentalQuadric = m.times(1/(float)count);
            applyQuadricToNeighbor();
        }

        public void unApplyQuadricToNeighbor() {
            if (nearest_in_other_mesh == null) return;
            if (fundamentalQuadric == null) return;
            nearest_in_other_mesh.unComputeError();
            nearest_in_other_mesh.quadric = nearest_in_other_mesh.quadric.minus(fundamentalQuadric);
            nearest_in_other_mesh.quadric_count--;
            if (nearest_in_other_mesh.quadric_count==0)
                nearest_in_other_mesh.quadric = Matrix.ZERO;
            nearest_in_other_mesh.computeError();
            nearest_in_other_mesh = null;
        }

        public void applyQuadricToNeighbor() {
            if (score_against == null) return;

            Vertex new_nearest = score_against.nearest(p);
            if (nearest_in_other_mesh != null && new_nearest == nearest_in_other_mesh) return;

            if (nearest_in_other_mesh != null) unApplyQuadricToNeighbor();
            if (nearest_in_other_mesh != null) throw new Error();

            nearest_in_other_mesh = new_nearest;
                
            // don't attract to vertices that face the other way
            if (nearest_in_other_mesh.e == null || nearest_in_other_mesh.norm().dot(norm()) < 0) {
                nearest_in_other_mesh = null;
            } else {
                nearest_in_other_mesh.unComputeError();
                nearest_in_other_mesh.quadric = nearest_in_other_mesh.quadric.plus(fundamentalQuadric());
                nearest_in_other_mesh.quadric_count++;
                nearest_in_other_mesh.computeError();
            }
            reComputeError();
        }

        public void reComputeErrorAround() {
            reComputeError();
            if (nearest_in_other_mesh != null) nearest_in_other_mesh.reComputeError();
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                e.p2.reComputeError();
        }
        public void reComputeError() {
            unComputeError();
            computeError();
        }
        public void unComputeError() {
            score -= oldscore;
            oldscore = 0;
        }
        public void computeError() {
            oldscore =
                quadric_count != 0
                ? (quadric.preAndPostMultiply(p) * 100) / quadric_count
                : immutableVertices
                ? oldscore
                : nearest_in_other_mesh != null
                ? nearest_in_other_mesh.fundamentalQuadric().preAndPostMultiply(p) * 100 * 10
                : score_against != null
                ? score_against.nearest(p).fundamentalQuadric().preAndPostMultiply(p) * 100 * 10
                : 0;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                double ang = Math.abs(e.crossAngle());
                if (ang > Math.PI) throw new Error();
                float minangle = (float)(Math.PI * 0.8);
                if (ang > minangle)
                    oldscore += (ang - minangle);
            }
            score += oldscore;
        }

        private void removeTrianglesFromRTree() {
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.t != null) e.t.removeFromRTree();
        }
        private void addTrianglesToRTree() {
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.t != null) e.t.addToRTree();
        }

        /** does NOT update bound pairs! */
        public boolean transform(Matrix m) {
            if (immutableVertices) throw new Error();

            unApplyQuadricToNeighbor();
            Point oldp = this.p;

            if (vertices.get(this.p)==null) throw new Error();
            vertices.remove(this);
            removeTrianglesFromRTree();
            float newx = m.a*p.x + m.b*p.y + m.c*p.z + m.d;
            float newy = m.e*p.x + m.f*p.y + m.g*p.z + m.h;
            float newz = m.i*p.x + m.j*p.y + m.k*p.z + m.l;
            this.p = new Point(newx, newy, newz);
            addTrianglesToRTree();
            vertices.add(this);

            applyQuadricToNeighbor();

            good = true;

            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next) {
                if (Math.abs(e.crossAngle()) > (Math.PI * 0.9) || Math.abs(e.next.crossAngle()) > (Math.PI * 0.9)) good = false;
                if (e.t.aspect() < 0.1) good = false;
                e.p2.quadricStale = true;
            }

            if (!ignorecollision && good) triangles.range(oldp, this.p, (Visitor<T>)this);

            reComputeErrorAround();
            return good;
        }

        public void visit(T t) {
            if (!good) return;
            for(E e = Vertex.this.e; e!=null; e=e.pair.next==Vertex.this.e?null:e.pair.next) {
                if (!t.has(e.p1) && !t.has(e.p2) && e.intersects(t)) { good = false; }
                if (e.t != null) {
                    if (!e.t.has(t.e1().p1) && !e.t.has(t.e1().p2) && t.e1().intersects(e.t)) { good = false; }
                    if (!e.t.has(t.e2().p1) && !e.t.has(t.e2().p2) && t.e2().intersects(e.t)) { good = false; }
                    if (!e.t.has(t.e3().p1) && !e.t.has(t.e3().p2) && t.e3().intersects(e.t)) { good = false; }
                }
            }
        }
        private boolean good;

        public boolean move(Vec v) {
            Matrix m = Matrix.translate(v);
            Vertex p = this;
            boolean good = true;
            do {
                good &= p.transform(m);
                p = p.bound_to;
            } while (p != this);
            return good;
        }

        public E getFreeIncident() {
            E ret = getFreeIncident(e, e);
            if (ret != null) return ret;
            ret = getFreeIncident(e.pair.next, e.pair.next);
            if (ret != null) return ret;
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                System.out.println(e + " " + e.t);
            throw new Error("unable to find free incident to " + this);
        }

        public E getFreeIncident(E start, E before) {
            for(E e = start; e!=null; e=e.pair.next==before?null:e.pair.next)
                if (e.pair.p2 == this && e.pair.t == null && e.pair.next.t == null)
                    return e.pair;
            return null;
        }

        public E getE(Point p2) {
            Vertex v = vertices.get(p2);
            if (v==null) return null;
            return getE(v);
        }
        public E getE(Vertex p2) {
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.p1 == this && e.p2 == p2) return e;
            return null;
        }

        public Vec norm() {
            Vec norm = new Vec(0, 0, 0);
            for(E e = this.e; e!=null; e=e.pair.next==this.e?null:e.pair.next)
                if (e.t != null)
                    norm = norm.plus(e.t.norm().times((float)e.prev.angle()));
            return norm.norm();
        }

        public boolean isBoundTo(Vertex p) {
            Vertex px = p;
            do {
                if (px==this) return true;
                px = px.bound_to;
            } while(px != p);
            return false;
        }
        public void unbind() { bound_to = this; binding = Matrix.ONE; }
        public void bind(Vertex p) { bind(p, Matrix.ONE); }
        public void bind(Vertex p, Matrix binding) {
            if (isBoundTo(p)) return;
            Vertex temp_bound_to = p.bound_to;
            Matrix temp_binding = p.binding;
            p.bound_to = this.bound_to;
            p.binding = binding.times(this.binding); // FIXME: may have order wrong here
            this.bound_to = temp_bound_to;
            this.binding = temp_binding.times(temp_binding); // FIXME: may have order wrong here
        }
    }

    public class BindingGroup {
        private HashSet<E> set = new HashSet<E>();
        public BindingGroup bind_others;
        public BindingGroup other() { return bind_others; }
        public BindingGroup(BindingGroup bind_others) { this.bind_others = bind_others; }
        public BindingGroup() { this.bind_others = new BindingGroup(this); }
        public BindingGroup(E e) { this(); set.add(e); }
        public void add(E e) {
            if (set.contains(e)) return;
            set.add(e);
            BindingGroup e_bind_peers = e.bind_peers;
            BindingGroup e_bind_to    = e.bind_to;
            e.bind_peers = this;
            e.bind_to    = bind_others;
            for (E epeer  : e_bind_peers.set) add(epeer);
            for (E eother : e_bind_to.set)    bind_others.add(eother);

            for(E eother : bind_others.set) {
                if (e.next.bind_to.set.contains(eother.prev)) {
                    e.next.next.bindEdge(eother.prev.prev);
                }
                if (e.prev.bind_to.set.contains(eother.next)) {
                    e.prev.prev.bindEdge(eother.next.next);
                }
            }

        }
        public void dobind(E e) {
            for(E ebound : set) {
                e.p1.bind(ebound.p2);
                e.p2.bind(ebound.p1);
            }
        }
        public void shatter(BindingGroup bg1, BindingGroup bg2) {
            for(E e : set) {
                e.shatter(e.midpoint(), bg1, bg2);
            }
        }
    }

    /** [UNIQUE] an edge */
    public final class E implements Comparable<E> {

        public final Vertex p1, p2;
        T t;     // triangle to our "left"
        E prev;  // previous half-edge
        E next;  // next half-edge
        E pair;  // partner half-edge
        public BindingGroup bind_peers  = new BindingGroup(this);
        public BindingGroup bind_to     = bind_peers.other();
        boolean shattered = false;

        public boolean intersects(T t) { return t.intersects(p1.p, p2.p); }
        public float comparator() {
            Vertex nearest = score_against.nearest(midpoint());
            //if (t==null) return length();
            /*
              double ang = Math.abs(crossAngle());
              float minangle = (float)(Math.PI * 0.9);
              if (ang > minangle)
              return 300;
            */
            /*
              if ((length() * length()) / t.area() > 10)
              return (float)(length()*Math.sqrt(t.area()));
              return length()*t.area();
            */
            return (float)Math.max(length(), midpoint().distance(nearest.p));
            //return length();
        }
        public int compareTo(E e) {
            return e.comparator() > comparator() ? 1 : -1;
        }
        public void bindEdge(E e) { bind_to.add(e); }
        public void dobind() { bind_to.dobind(this); }

        public Point shatter() { return shatter(midpoint(), null, null); }
        public Point shatter(Point mid, BindingGroup bg1, BindingGroup bg2) {
            if (shattered || destroyed) return mid;
            shattered = true;

            Vertex r = next.p2;
            E next = this.next;
            E prev = this.prev;

            int old_colorclass = t==null ? 0 : t.colorclass;
            if (bg1==null) bg1 = new BindingGroup();
            if (bg2==null) bg2 = new BindingGroup();
            BindingGroup old_bind_to = bind_to;
            bind_peers.shatter(bg1, bg2);
            old_bind_to.shatter(bg2.other(), bg1.other());
            pair.shatter();
            destroy();

            newT(r.p, p1.p, mid, null, old_colorclass);
            newT(r.p, mid, p2.p, null, old_colorclass);
            bg1.add(p1.getE(mid));
            bg2.add(p2.getE(mid).pair);
            return mid;
        }

        public boolean destroyed = false;
        public void destroy() {
            if (destroyed) return;
            destroyed = true;
            pair.destroyed = true;

            if (t != null) t.destroy();
            t = null;

            if (pair.t != null) pair.t.destroy();
            pair.t = null;

            if (next.t != null) next.t.destroy();
            if (prev.t != null) prev.t.destroy();
            next.t = null;
            prev.t = null;

            if (pair.next.t != null) pair.next.t.destroy();
            if (pair.prev.t != null) pair.next.t.destroy();
            pair.next.t = null;
            pair.prev.t = null;

            this.bind_to = null;
            pair.bind_to = null;
            this.bind_peers = null;
            pair.bind_peers = null;
            pair.prev.next = next;
            next.prev = pair.prev;
            prev.next = pair.next;
            pair.next = prev;
            if (p1.e == this) p1.e = prev.next;
            if (pair.p1.e == pair) pair.p1.e = pair.prev.next;
        }

        private void sync() {
            this.prev.next = this;
            this.next.prev = this;
            this.pair.pair = this;
            bind_peers.add(this);
            if (this.next.p1 != p2) throw new Error();
            if (this.prev.p2 != p1) throw new Error();
            if (this.p1.e == null) this.p1.e = this;
            if (!added) added = true;
        }
        private boolean added = false;

        public T makeT(int colorclass) { return t==null ? (t = new T(this, colorclass)) : t; }

        public double crossAngle() {
            Vec v1 = t.norm().times(-1);
            Vec v2 = pair.t.norm().times(-1);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        /** angle between this half-edge and the next */
        public double angle() {
            Vec v1 = next.p2.p.minus(p2.p);
            Vec v2 = this.p1.p.minus(p2.p);
            return Math.acos(v1.norm().dot(v2.norm()));
        }

        public void makeAdjacent(E e) {
            if (this.next == e) return;
            if (p2 != e.p1) throw new Error("cannot make adjacent -- no shared vertex");
            if (t != null || e.t != null) throw new Error("cannot make adjacent -- edges not both free");

            E freeIncident = p2.getFreeIncident(e, this);

            e.prev.next = freeIncident.next;
            freeIncident.next.prev = e.prev;

            freeIncident.next = this.next;
            this.next.prev = freeIncident;
            
            this.next = e;
            e.prev = this;

            sync();
            freeIncident.sync();
        }

        /** creates an isolated edge out in the middle of space */
        public E(Point p1, Point p2) {
            if (vertices.get(p1) != null) throw new Error();
            if (vertices.get(p2) != null) throw new Error();
            this.p1 = new Vertex(p1);
            this.p2 = new Vertex(p2);
            this.prev = this.next = this.pair = new E(this, this, this);
            this.p1.e = this;
            this.p2.e = this.pair;
            sync();
        }

        /** adds a new half-edge from prev.p2 to p2 */
        public E(E prev, Point p) {
            Vertex p2;
            p2 = vertices.get(p);
            if (p2 == null) p2 = new Vertex(p);
            this.p1 = prev.p2;
            this.p2 = p2;
            this.prev = prev;
            if (p2.getE(p1) != null) throw new Error();
            if (p2.e==null) {
                this.next = this.pair = new E(this, this, prev.next);
            } else {
                E q = p2.getFreeIncident();
                this.next = q.next;
                this.next.prev = this;
                E z = prev.next;
                this.prev.next = this;
                this.pair = new E(q, this, z);
            }
            if (p2.e==null) p2.e = this.pair;
            sync();
        }

        /** adds a new half-edge to the mesh with a given predecessor, successor, and pair */
        public E(E prev, E pair, E next) {
            this.p1 = prev.p2;
            this.p2 = next.p1;
            this.prev = prev;
            this.next = next;
            this.pair = pair;
            sync();
        }
        public Point midpoint() { return new Point((p1.p.x+p2.p.x)/2, (p1.p.y+p2.p.y)/2, (p1.p.z+p2.p.z)/2); }
        public boolean has(Vertex v) { return v==p1 || v==p2; }
        public float length() { return p1.p.minus(p2.p).mag(); }
        public String toString() { return p1+"->"+p2; }

    }

    public E makeE(Point p1, Point p2) {
        Vertex v1 = vertices.get(p1);
        Vertex v2 = vertices.get(p2);
        if (v1 != null && v2 != null) {
            E e = v1.getE(v2);
            if (e != null) return e;
            e = v2.getE(v1);
            if (e != null) return e;
        }
        if (v1 != null) return new E(v1.getFreeIncident(), p2);
        if (v2 != null) return new E(v2.getFreeIncident(), p1).pair;
        return new E(p1, p2);
    }
    public T newT(Point p1, Point p2, Point p3, Vec norm, int colorclass) {
        if (norm != null) {
            Vec norm2 = p3.minus(p1).cross(p2.minus(p1));
            float dot = norm.dot(norm2);
            //if (Math.abs(dot) < EPointSILON) throw new Error("dot products within evertsilon of each other: "+norm+" "+norm2);
            if (dot < 0) { Point p = p1; p1=p2; p2 = p; }
        }
        E e12 = makeE(p1, p2);
        E e23 = makeE(p2, p3);
        E e31 = makeE(p3, p1);
        while(e12.next != e23 || e23.next != e31 || e31.next != e12) {
            e12.makeAdjacent(e23);
            e23.makeAdjacent(e31);
            e31.makeAdjacent(e12);
        }
        T ret = e12.makeT(colorclass);
        if (e12.t == null) throw new Error();
        if (e23.t == null) throw new Error();
        if (e31.t == null) throw new Error();
        return ret;
    }


    /** [UNIQUE] a triangle (face) */
    public final class T extends Triangle {
        public final E e1;
        public final int color;
        public final int colorclass;

        public void removeFromRTree() { triangles.remove(this); }
        public void addToRTree() { triangles.insert(this); }

        public void destroy() { triangles.remove(this); }

        T(E e1, int colorclass) {
            this.e1 = e1;
            E e2 = e1.next;
            E e3 = e2.next;
            if (e1==e2 || e1==e3) throw new Error();
            if (e3.next!=e1) throw new Error();
            if (e1.t!=null || e2.t!=null || e3.t!=null) throw new Error("non-manifold surface or disagreeing normals");
            e1.t = this;
            e1.next.t = this;
            e1.next.next.t = this;

            // FIXME: check for sealed/watertight surface once construction is complete (and infer normal(s)?)

            int color = Math.abs(random.nextInt());
            while(true) {
                color = color % 4;
                if (e1().pair.t != null && color == e1().pair.t.color) { color++; continue; }
                if (e2().pair.t != null && color == e2().pair.t.color) { color++; continue; }
                if (e3().pair.t != null && color == e3().pair.t.color) { color++; continue; }
                break;
            }
            this.color = color;
            this.colorclass = colorclass;
            triangles.add(this);
        }
        public E e1() { return e1; }
        public E e2() { return e1.next; }
        public E e3() { return e1.prev; }
        public Vertex v1() { return e1.p1; }
        public Vertex v2() { return e1.p2; }
        public Vertex v3() { return e1.next.p2; }
        public Point p1() { return e1.p1.p; }
        public Point p2() { return e1.p2.p; }
        public Point p3() { return e1.next.p2.p; }
        public boolean hasE(E e) { return e1==e || e1.next==e || e1.prev==e; }
        public boolean has(Vertex v) { return v1()==v || v2()==v || v3()==v; }

        public boolean shouldBeDrawn() {
            if (e1().bind_to.set.size() == 0) return false;
            if (e2().bind_to.set.size() == 0) return false;
            if (e3().bind_to.set.size() == 0) return false;
            return true;
        }

    }
}
